/* Copyright (C) 2000 Ross Diankov

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

// Name:	math3d.cpp
// Purpose: Provides a basic structure for the mathematics involved in the
//			3d pipeline and functions that manipulate matrices and vectors.

#include <math.h>
#include <memory.h>

#include "math3d.h"

///////////////////////
// Matrix Operations //
///////////////////////

void MatrixIdentity(MATRIX&a)
{
	memset(&a, 0, sizeof(a));

	a._11 = 1.0f;
	a._22 = 1.0f;
	a._33 = 1.0f;
	a._44 = 1.0f;
}

// the inverse of a matrix is devined by A-1 = CT/det
// where CT is the transposed coefficient matrix and 
// det is the determinant of A
// Note: This word only for matrices where the 4th column is [0 0 0 1)
void MatrixInvert(MATRIX& q, MATRIX& a)
{
    SCALAR fDetInv = 1.0f / ( a(0, 0) * ( a(1, 1) * a(2, 2) - a(1, 2) * a(2, 1) ) -
                              a(0, 1) * ( a(1, 0) * a(2, 2) - a(1, 2) * a(2, 0) ) +
                              a(0, 2) * ( a(1, 0) * a(2, 1) - a(1, 1) * a(2, 0) ) );
 
    q(0, 0) =  fDetInv * ( a(1, 1) * a(2, 2) - a(1, 2) * a(2, 1) );
    q(0, 1) = -fDetInv * ( a(0, 1) * a(2, 2) - a(0, 2) * a(2, 1) );
    q(0, 2) =  fDetInv * ( a(0, 1) * a(1, 2) - a(0, 2) * a(1, 1) );
    q(0, 3) = 0.0f;

    q(1, 0) = -fDetInv * ( a(1, 0) * a(2, 2) - a(1, 2) * a(2, 0) );
    q(1, 1) =  fDetInv * ( a(0, 0) * a(2, 2) - a(0, 2) * a(2, 0) );
    q(1, 2) = -fDetInv * ( a(0, 0) * a(1, 2) - a(0, 2) * a(1, 0) );
    q(1, 3) = 0.0f;

    q(2, 0) =  fDetInv * ( a(1, 0) * a(2, 1) - a(1, 1) * a(2, 0) );
    q(2, 1) = -fDetInv * ( a(0, 0) * a(2, 1) - a(0, 1) * a(2, 0) );
    q(2, 2) =  fDetInv * ( a(0, 0) * a(1, 1) - a(0, 1) * a(1, 0) );
    q(2, 3) = 0.0f;

    q(3, 0) = -( a(3, 0) * q(0, 0) + a(3, 1) * q(1, 0) + a(3, 2) * q(2, 0) );
    q(3, 1) = -( a(3, 0) * q(0, 1) + a(3, 1) * q(1, 1) + a(3, 2) * q(2, 1) );
    q(3, 2) = -( a(3, 0) * q(0, 2) + a(3, 1) * q(1, 2) + a(3, 2) * q(2, 2) );
    q(3, 3) = 1.0f;
}

void MatrixMultiply(MATRIX& a, MATRIX& b, MATRIX& c)
{	
	MATRIX q;
	for(int i = 0; i < 4; i++) {
		for(int j = 0; j < 4; j++) { 
			q(i, j) = b(i, 0) * c(0, j) + 
					  b(i, 1) * c(1, j) +
					  b(i, 2) * c(2, j) +
					  b(i, 3) * c(3, j);
		}
	}
		
	a = q;
}

void VectorMatrixMultiply(VECTOR& ans, MATRIX& trans, VECTOR& a)
{
	VECTOR v;
	SCALAR w;					// for homogeneous coordinates divide by w

	v.x = a.x * trans._11 + a.y * trans._21 + a.z * trans._31 + trans._41;
	v.y = a.x * trans._12 + a.y * trans._22 + a.z * trans._32 + trans._42;
	v.z = a.x * trans._13 + a.y * trans._23 + a.z * trans._33 + trans._43;
	w	= a.x * trans._14 + a.y * trans._24 + a.z * trans._34 + trans._44;

	ans = v / w;				// homogenize
}


////////////////////////////
// Matrix Transformations //
////////////////////////////

void SetMatrixRotateX(MATRIX& a, SCALAR fAng)
{
	SCALAR fCos = cosf(fAng);
	SCALAR fSin = sinf(fAng);

	MatrixIdentity(a);
	a(1, 1) =  fCos;
	a(2, 1) = -fSin;
	a(1, 2) =  fSin;
	a(2, 2) =  fCos;
}

void SetMatrixRotateY(MATRIX& a, SCALAR fAng)
{
	SCALAR fCos = cosf(fAng);
	SCALAR fSin = sinf(fAng);

	MatrixIdentity(a);
	a(0, 0) =  fCos;
	a(2, 0) =  fSin;
	a(0, 2) = -fSin;
	a(2, 2) =  fCos;
}

void SetMatrixRotateZ(MATRIX& a, SCALAR fAng)
{
	SCALAR fCos = cosf(fAng);
	SCALAR fSin = sinf(fAng);

	MatrixIdentity(a);
	a(0, 0) =  fCos;
	a(1, 0) = -fSin;
	a(0, 1) =  fSin;
	a(1, 1) =  fCos;
}

void SetMatrixTranslate(MATRIX &a, VECTOR& dir)
{
	MatrixIdentity(a);
	a(3, 0) = dir.x;
	a(3, 1) = dir.y;
	a(3, 2) = dir.z;
}

void SetMatrixTranslate(MATRIX& a, SCALAR tx, SCALAR ty, SCALAR tz)
{
	MatrixIdentity(a);
	a(3, 0) = tx;
	a(3, 1) = ty;
	a(3, 2) = tz;
}

void SetMatrixScale(MATRIX&a, SCALAR scale)
{
	MatrixIdentity(a);
	a(0, 0) = scale;
	a(1, 1) = scale;
	a(2, 2) = scale;
}

///////////////////////
// Vector Operations //
///////////////////////

SCALAR Magnitude(VECTOR& a)
{
	return sqrtf(a.x * a.x + a.y * a.y + a.z * a.z);
}

SCALAR DotProduct(VECTOR& a, VECTOR& b)
{
	return ((a.x * b.x) + (a.y * b.y) + (a.z * b.z));
}

VECTOR CrossProduct(VECTOR& a, VECTOR& b)
{
	return VECTOR(a.y * b.z - a.z * b.y, 
				  a.z * b.x - a.x * b.z, 
				  a.x * b.y - a.y * b.x);
}

VECTOR Normalize(VECTOR& a)
{
	SCALAR fLength = 1.0f / sqrtf((a.x * a.x) + (a.y * a.y) + (a.z * a.z));

	a = a * fLength;
	return a;
}