/* Copyright (C) 2000 Ross Diankov

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

// Name:	math3d.h
// Purpose: Provides a basic structure for the mathematics involved in the
//			3d pipeline and functions that manipulate matrices and vectors.

#ifndef VECTORMATRIX_MATH
#define VECTORMATRIX_MATH

#define G_PI				3.1415926f
#define G_EPSILON			1.0e-4f

#define max(a,b) (((a) > (b)) ? (a) : (b))
#define min(a,b) (((a) < (b)) ? (a) : (b))

typedef float SCALAR;

struct VECTOR
{

	SCALAR x, y, z; //x, y, and z homogeneous coordinates

	VECTOR() : x(0), y(0), z(0) {}
	VECTOR( const SCALAR& a, const SCALAR& b, const SCALAR& c )
		: x(a), y(b), z(c) {}

	//index a component
	//NOTE: returning a reference allows
	//you to assign the indexed element
	SCALAR& operator ()( const long i ) { return *((&x) + i); }
	
	//negate
	const VECTOR operator - () const { return VECTOR( -x, -y, -z ); }

	//assign
	const VECTOR& operator = ( const VECTOR& v ) 
	{
		memcpy(this, &v, sizeof(VECTOR));
		return *this;
	}

	//add
	const VECTOR operator + ( const VECTOR& v ) const
	{ 
		return VECTOR(x + v.x, y + v.y, z + v.z); 
	}

	//subtract
	const VECTOR operator - ( const VECTOR& v ) const
	{ 
		return VECTOR(x - v.x, y - v.y, z - v.z); 
	}

	//post-multiply by a scalar
	const VECTOR operator * ( const SCALAR& s ) const
	{ 
		return VECTOR( x*s, y*s, z*s ); 
	}

	const VECTOR operator / (SCALAR s) const 
	{ 
		s = 1/s;
		return VECTOR( s*x, s*y, s*z );

	}

	const bool operator == ( const VECTOR& v) const
	{
		if(fabs(v.x - x) < G_EPSILON && 
			fabs(v.y - y) < G_EPSILON && 
			fabs(v.z - z) < G_EPSILON) return true;
		return false;
	}

	const bool operator != ( const VECTOR& v) const
	{
		if( *this == v ) return false;
		return true;
	}
};

struct MATRIX
{
	SCALAR _11,	_12, _13, _14;
	SCALAR _21, _22, _23, _24;
	SCALAR _31, _32, _33, _34;
	SCALAR _41, _42, _43, _44;

	SCALAR& operator ()(const unsigned int i, const unsigned int j) 
	{
		return *(SCALAR*)((long)this + (i * sizeof(SCALAR) * 4) + (j * sizeof(SCALAR)));
	}
};

// Matrix operations
void MatrixIdentity(MATRIX&);
void MatrixInvert(MATRIX&, MATRIX&);
void MatrixMultiply(MATRIX&, MATRIX&, MATRIX&);
void VectorMatrixMultiply(VECTOR&, MATRIX&, VECTOR&);

// 3D transformation, angle has to be in radians
void SetMatrixRotateX(MATRIX& a, SCALAR);
void SetMatrixRotateY(MATRIX& a, SCALAR);
void SetMatrixRotateZ(MATRIX& a, SCALAR);
void SetMatrixTranslate(MATRIX&, VECTOR& t);
void SetMatrixTranslate(MATRIX&, SCALAR tx, SCALAR ty, SCALAR tz);
void SetMatrixScale(MATRIX&, SCALAR);

// Vector operations
SCALAR Magnitude(VECTOR&);
SCALAR DotProduct(VECTOR&, VECTOR&);
VECTOR CrossProduct(VECTOR&, VECTOR&);
VECTOR Normalize(VECTOR&);

#endif