/* Copyright (C) 2000 Ross Diankov

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

// Name:	object3d.cpp
// Purpose: Organizes 3d data in structures and provides the basic
//          functions to initiate and transform 3d objects.
 
#include <windows.h>
#include <stdio.h>
#include <math.h>

#include "math3d.h"
#include "object3d.h"

#define MSG(x) MessageBox(NULL, x, "Application message", MB_OK);


// VERTEX3D constructor
VERTEX3D::VERTEX3D(VECTOR& Pos, VECTOR &norm, SCALAR tu, SCALAR tv)
{
	pos = Pos;
	normal = norm;
	u = tu;
	v = tv;
	color = 0.0f;
}

// Object3D constructor
OBJECT3D::OBJECT3D(int numverts, int numfaces)
{
	faces = 0;
	face_normals = 0;
	vertices = 0;

	numVertices = numverts;
	numFaces = numfaces;

	// alocate memory for vertex and face info
	vertices = new VERTEX3D[numVertices];
	
	if(numfaces) faces = new WORD[numFaces * 3];
	if(numfaces) face_normals = new VECTOR[numFaces];
}

// SetProjectionMatrix()
// Sets a projection matrix with field of views and near and far clipping planes.
// fFOVx is the x-axis field of view divided by 2
// fFOVy is the y-axis field of view divided by 2
// fNearPlane and fFarPlane are the near and far clipping planes
// fAspect is the ratio of Height to Width
void SetProjectionMatrix(MATRIX& proj, SCALAR fFOVx, SCALAR fFOVy, SCALAR fAspect,
						 SCALAR fNearPlane, SCALAR fFarPlane)
{
	SCALAR w = fAspect * cosf(fFOVx) / sinf(fFOVx);
	SCALAR h = cosf(fFOVy) / sinf(fFOVy);
	SCALAR fPlane = fFarPlane / (fFarPlane - fNearPlane);

	MatrixIdentity(proj);
	proj._11 = w;
	proj._22 = h;
	proj._33 = fPlane;
	proj._34 = 1.0f;
	proj._43 = -fPlane * fNearPlane;
	proj._44 = 0.0f;
}

// Interpolates a two vectors by a scalar value
void VectorInterpolate(VECTOR& interp, VECTOR& start, VECTOR& end, SCALAR t)
{
	interp.x = start.x + (end.x - start.x) * t;
	interp.y = start.y + (end.y - start.y) * t;
	interp.z = start.z + (end.z - start.z) * t;
}

// compute the normals for every face in the object
// you would only need two vectors  v and w
void ComputeObjectNormals(OBJECT3D* obj)
{
	for(int i = 0; i < obj->numFaces; i++) {
		VECTOR v = obj->vertices[obj->faces[i*3+2]].pos - obj->vertices[obj->faces[i*3]].pos;
		VECTOR w = obj->vertices[obj->faces[i*3+1]].pos - obj->vertices[obj->faces[i*3]].pos;
		obj->face_normals[i] = Normalize(CrossProduct(v, w));
	}
}

// Create a Rotation matrix about vector direction
// copyright by DirectX
void SetRotationMatrix( MATRIX& mat, VECTOR& vDir, SCALAR fRads )
{
    SCALAR fCos = cosf(fRads);
    SCALAR fSin = sinf(fRads);
    VECTOR v    = Normalize( vDir );

    mat._11 = ( v.x * v.x ) * ( 1.0f - fCos ) + fCos;
    mat._12 = ( v.x * v.y ) * ( 1.0f - fCos ) - (v.z * fSin);
    mat._13 = ( v.x * v.z ) * ( 1.0f - fCos ) + (v.y * fSin);

    mat._21 = ( v.y * v.x ) * ( 1.0f - fCos ) + (v.z * fSin);
    mat._22 = ( v.y * v.y ) * ( 1.0f - fCos ) + fCos ;
    mat._23 = ( v.y * v.z ) * ( 1.0f - fCos ) - (v.x * fSin);

    mat._31 = ( v.z * v.x ) * ( 1.0f - fCos ) - (v.y * fSin);
    mat._32 = ( v.z * v.y ) * ( 1.0f - fCos ) + (v.x * fSin);
    mat._33 = ( v.z * v.z ) * ( 1.0f - fCos ) + fCos;
    
    mat._14 = mat._24 = mat._34 = 0.0f;
    mat._41 = mat._42 = mat._43 = 0.0f;
    mat._44 = 1.0f;
} 