/* Copyright (C) 2000 Ross Diankov

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <windows.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "math3d.h"
#include "object3d.h"
#include "device3d.h"

#include "resource.h"

#define SCREEN_WIDTH		640
#define SCREEN_HEIGHT		480
#define WINDOW_CLASS_NAME	"Windows Re-use class"

#define SAFE_DELETE(x) if(x) { delete x; x = NULL; }
#define MSG(x) MessageBox(NULL, x, "Application Message", MB_OK)

HWND hWnd;					// global variable that stores the handle to window
RECT rcWindow;				// the window's dimensions
int wndWidth, wndHeight;
BOOL f_Active = FALSE;

Device3D *d3dDevice;
LIGHT3D* lightDir;
LIGHT3D* lightAmb;
MATERIAL3D* m3dMat;

OBJECT3D *pyramid;
OBJECT3D *plane;

MATRIX objTrans, newrot;
SCALAR fAng = 0.03f;
bool left_drag = FALSE;
bool right_drag = FALSE;
POINT ptPrev;

VECTOR vPos, vLookAt, vUp;

int GetMenuShadeModePos(DWORD mode);
void SetMenuShadeMode(Device3D*, DWORD mode);
COLOR3D GetNewColor(MATERIAL3D* curr);

BOOL InitObjects();
void RenderObjects();
void MoveObjects();
void ShowStats();
void DestroyObjects();

LRESULT CALLBACK AboutDlgProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK ColorDlgProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK WinProc(HWND, UINT, WPARAM, LPARAM); // main window proc

int WINAPI WinMain (HINSTANCE hinstance, HINSTANCE, LPSTR, int howshow)
{
	MSG         msg;
	WNDCLASS wndclass;

	wndclass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wndclass.lpfnWndProc = WinProc;
	wndclass.cbClsExtra    = 0;
	wndclass.cbWndExtra    = 0;
	wndclass.hInstance     = hinstance;
	wndclass.hIcon         = LoadIcon (NULL, "ENG3D_ICON");
	wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW);
	wndclass.hbrBackground = (HBRUSH)GetStockObject (WHITE_BRUSH);
	wndclass.lpszMenuName  = "ENGINE3DMENU";
	wndclass.lpszClassName = WINDOW_CLASS_NAME;

	if(!RegisterClass (&wndclass)) return FALSE;

	hWnd = CreateWindow (WINDOW_CLASS_NAME, "Windows Application", WS_OVERLAPPEDWINDOW,
							 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT,
							 NULL, NULL, hinstance, NULL);

	if (!hWnd) return 0L;                    // window construction failed

	ShowWindow (hWnd, howshow); // creates window on the screen
	UpdateWindow (hWnd);        // paints the client (user area) of the window
	
	GetClientRect(hWnd, &rcWindow);
	wndWidth = rcWindow.right - rcWindow.left;
	wndHeight = rcWindow.bottom - rcWindow.top;

	if(!InitObjects()) return 0L;

	f_Active = TRUE;

	while (1) {
		if(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
			if(msg.message == WM_QUIT) break;

			TranslateMessage (&msg);
			DispatchMessage  (&msg);
		}
	
		MoveObjects();
		RenderObjects();
	}

	f_Active = FALSE;
	DestroyObjects();

	return msg.wParam;
}

// Proccessing message functions

LRESULT CALLBACK AboutDlgProc(HWND hDlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    switch (msg) {
		case WM_INITDIALOG:
			return 1L;

		case WM_COMMAND:
			if (wparam == IDOK) {   
				EndDialog(hDlg, TRUE);
				return 1L;
			}
			break;
    }

    return 0L;
}

LRESULT CALLBACK ColorDlgProc(HWND hDlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
	COLOR3D* col;
	char str[10];

	switch(msg) {
		case WM_INITDIALOG:
			col = (COLOR3D*)GetWindowLong(GetParent(hDlg), GWL_USERDATA);
			
			sprintf(str, "%d", (int)(col->r * 255.0f));
			SetWindowText(GetDlgItem(hDlg, IDC_REDCOLOR), str);
			sprintf(str, "%d", (int)(col->g * 255.0f));
			SetWindowText(GetDlgItem(hDlg, IDC_GREENCOLOR), str);
			sprintf(str, "%d", (int)(col->b * 255.0f));
			SetWindowText(GetDlgItem(hDlg, IDC_BLUECOLOR), str);

			return 1L;

		case WM_COMMAND: 
			if(wparam == IDOK) {
				col = (COLOR3D*)GetWindowLong(GetParent(hDlg), GWL_USERDATA);
				GetDlgItemText(hDlg, IDC_REDCOLOR, str, 4);
				col->r = (float)max(min(atoi(str), 255), 0) / 255.0f;
				GetDlgItemText(hDlg, IDC_GREENCOLOR, str, 4);
				col->g = (float)max(min(atoi(str), 255), 0) / 255.0f;
				GetDlgItemText(hDlg, IDC_BLUECOLOR, str, 4);
				col->b = (float)max(min(atoi(str), 255), 0) / 255.0f;

				EndDialog(hDlg, TRUE);
			}
			break;
	}

	return 0L;
}

LRESULT CALLBACK WinProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	HDC         hdc;
	PAINTSTRUCT ps;
	MATRIX mat;
	VECTOR vVector;

	switch (msg) {
		case WM_PAINT: 
			hdc = BeginPaint (hwnd, &ps);

			EndPaint (hwnd, &ps);
			break;
			
		case WM_DESTROY:
			PostQuitMessage (0);
			break;

		case WM_SIZE:
			if(f_Active) {
				GetClientRect(hWnd, &rcWindow);
				if(!d3dDevice->Update()) return 0L;
				break;
			}

		case WM_LBUTTONDOWN:
			if(f_Active) {
				if(TestPointInObject(d3dDevice, pyramid, LOWORD(lparam), HIWORD(lparam))) {
					
					SetCapture(hWnd);
					left_drag = TRUE;
					ptPrev.x = LOWORD(lparam);
					ptPrev.y = HIWORD(lparam);
					MatrixIdentity(newrot);
				}
			}
			break;

		case WM_LBUTTONUP:
			if(f_Active && left_drag) {
				ReleaseCapture();
				left_drag = FALSE;
				MatrixMultiply(objTrans, objTrans, newrot);
			}
			break;

		case WM_MOUSEMOVE:
			if(f_Active) {
				if( (wparam & MK_LBUTTON) && left_drag) {
					VECTOR vRot;
					SCALAR deltax = (SCALAR)LOWORD(lparam) - (SCALAR)ptPrev.x;
					SCALAR deltay = (SCALAR)HIWORD(lparam) - (SCALAR)ptPrev.y;
					SCALAR dist = sqrtf(deltax * deltax + deltay * deltay);

					vRot.x = deltay / dist;
					vRot.y = deltax / dist;
					vRot.z = 0.0f;

					SetRotationMatrix(newrot, vRot, dist * 0.01f);
				}
			}
			break;

		case WM_COMMAND: 
			switch(LOWORD(wparam)) {
				case MENU_HELP_ABOUT:
					DialogBox((HINSTANCE)GetWindowLong(hWnd, GWL_HINSTANCE), "IDD_ENG3DABOUT", hWnd, (DLGPROC)AboutDlgProc);
					break;

				case MENU_RENDER_WIREFRAME:
					SetMenuShadeMode(d3dDevice, SHADE_WIREFRAME);
					break;

				case MENU_RENDER_FLAT:
					SetMenuShadeMode(d3dDevice, SHADE_FLAT);
					break;

				case MENU_RENDER_GOURAUD:
					SetMenuShadeMode(d3dDevice, SHADE_GOURAUD);
					break;

				case MENU_RENDER_CHANGECOLOR:
					COLOR3D newcol = GetNewColor(m3dMat);

					m3dMat->diffuse.r = newcol.r;
					m3dMat->diffuse.g = newcol.g;
					m3dMat->diffuse.b = newcol.b;

					d3dDevice->SetMaterial(m3dMat);
					break;

				case MENU_FILE_EXIT:
					SendMessage(hWnd, WM_DESTROY, 0L, 0L);
					break;

				case MENU_FILE_RESET:
					MatrixIdentity(objTrans);
					MatrixIdentity(newrot);
					break;

				
			}
			break;

		case WM_KEYDOWN:
			switch((int)wparam) {
				case 'A':					// a
					vVector = Normalize((VECTOR)(vLookAt - vPos));
					vPos = vPos + (vVector * 4.0f);
					d3dDevice->SetViewMatrix(vPos, vLookAt, vUp);
					break;

				case 'Z':					// z
					vVector = Normalize((VECTOR)(vLookAt - vPos));
					vPos = vPos + (vVector * -4.0f);
					d3dDevice->SetViewMatrix(vPos, vLookAt, vUp);
					break;

				case VK_LEFT:					// left arrow
					SetMatrixRotateY(mat, -0.05f);
					VectorMatrixMultiply(vVector, mat, (VECTOR)(vLookAt - vPos));
					vLookAt = vPos + vVector;
					d3dDevice->SetViewMatrix(vPos, vLookAt, vUp);
					break;
					
				case VK_RIGHT:					// right arrow
					SetMatrixRotateY(mat, 0.05f);
					VectorMatrixMultiply(vVector, mat, (VECTOR)(vLookAt - vPos));
					vLookAt = vPos + vVector;
					d3dDevice->SetViewMatrix(vPos, vLookAt, vUp);
					break;

				case VK_UP:					// up arrow
					SetMatrixRotateX(mat, 0.05f);
					VectorMatrixMultiply(vVector, mat, (VECTOR)(vLookAt - vPos));
					vLookAt = vPos + vVector;
					d3dDevice->SetViewMatrix(vPos, vLookAt, vUp);
					break;

				case VK_DOWN:					// down arrow
					SetMatrixRotateX(mat, -0.05f);
					VectorMatrixMultiply(vVector, mat, (VECTOR)(vLookAt - vPos));
					vLookAt = vPos + vVector;
					d3dDevice->SetViewMatrix(vPos, vLookAt, vUp);
					break;

				case 'X':					// x
					d3dDevice->GetViewMatrix(mat);
					vPos = vPos - VECTOR(mat._11, mat._21, mat._31) * 5.0f;
					vLookAt = vLookAt - VECTOR(mat._11, mat._21, mat._31) * 5.0f;
					d3dDevice->SetViewMatrix(vPos, vLookAt, vUp);
					break;

				case 'C':					// c
					d3dDevice->GetViewMatrix(mat);
					vPos = vPos + VECTOR(mat._11, mat._21, mat._31) * 5.0f;
					vLookAt = vLookAt + VECTOR(mat._11, mat._21, mat._31) * 5.0f;
					d3dDevice->SetViewMatrix(vPos, vLookAt, vUp);
					break;
	
			}
			break;
	}

	return DefWindowProc (hwnd, msg, wparam, lparam);
}

COLOR3D GetNewColor(MATERIAL3D* curr)
{
	COLOR3D col;

	col.r = curr->diffuse.r;
	col.g = curr->diffuse.g;
	col.b = curr->diffuse.b;

	SetWindowLong(hWnd, GWL_USERDATA, (long)&col);

	DialogBox((HINSTANCE)GetWindowLong(hWnd, GWL_HINSTANCE), "IDD_COLORDLG", hWnd, (DLGPROC)ColorDlgProc);

	return col;
}

int GetMenuShadeModePos(DWORD mode)
{
	switch(mode) {
		case SHADE_WIREFRAME:	return 0;
		case SHADE_FLAT:		return 1;
		case SHADE_GOURAUD:		return 2;
	}

	return 0;
}

void SetMenuShadeMode(Device3D* dev, DWORD mode)
{
	HMENU menu;
	DWORD oldmode;

	menu = GetMenu(hWnd);
	menu = GetSubMenu(menu, 1);

	dev->GetShadingMode(oldmode);

	CheckMenuItem(menu, GetMenuShadeModePos(oldmode), MF_BYPOSITION | MF_UNCHECKED);
	CheckMenuItem(menu, GetMenuShadeModePos(mode), MF_BYPOSITION | MF_CHECKED);

	dev->SetShadingMode(mode);
}

BOOL InitObjects()
{
	VECTOR norm(0.0f, 0.0f, 0.0f);
	MATRIX world, proj;
	SCALAR f45DegAng = sinf(G_PI * 0.25f);
	
	d3dDevice = new Device3D();

	if(!d3dDevice->Init(hWnd)) {
		MSG("Error occured while intiating the device");
		return TRUE;
	}

	vPos = VECTOR(0.0f, 0.0f, -200.0f);
	vLookAt = VECTOR(0.0f, 0.0f, 0.0f);
	vUp = VECTOR(0.0f, 1.0f, 0.0f);

	MatrixIdentity(world);
	SetProjectionMatrix(proj, G_PI * 0.25f, G_PI * 0.25f, 1.0f, 1.0f, 1000.0f);

	d3dDevice->SetWorldMatrix(world);
	d3dDevice->SetViewMatrix(vPos, vLookAt, vUp);
	d3dDevice->SetProjectMatrix(proj);	

	lightDir = new LIGHT3D;
	lightDir->dir = VECTOR(0.0f, 1.0f, -1.0f);
	lightDir->color.r = 1.0f;
	lightDir->color.g = 1.0f;
	lightDir->color.b = 1.0f;
	Normalize(lightDir->dir);
	
	lightAmb = new LIGHT3D;
	lightAmb->color.r = 0.1f;
	lightAmb->color.g = 0.1f;
	lightAmb->color.b = 0.1f;

	m3dMat = new MATERIAL3D;
	m3dMat->diffuse.r = 0.2f;
	m3dMat->diffuse.g = 0.95f;
	m3dMat->diffuse.b = 0.2f;
	m3dMat->ambient.r = 1.0f;
	m3dMat->ambient.g = 1.0f;
	m3dMat->ambient.b = 1.0f;

	d3dDevice->SetLight(LIGHT_AMBIENT, lightAmb);
	d3dDevice->SetLight(LIGHT_DIRECTIONAL, lightDir);
	d3dDevice->SetMaterial(m3dMat);

	SetMenuShadeMode(d3dDevice, SHADE_FLAT);
	d3dDevice->SetRenderMode(RENDER_FILLFACE);

	pyramid = new OBJECT3D(6, 8);

	//       *4
	//      / \
	//     /   \
	//	  //   \\
	//   1*------*3
	//  0*\_____/*2
	//	 \ \   / /
	//	  \     /
	//     \   /
	//       *5
	pyramid->vertices[0] = VERTEX3D(VECTOR(-50.0f, 0.0f, -50.0f), VECTOR(-f45DegAng, 0.0f, -f45DegAng), 0.0f, 0.0f);
	pyramid->vertices[1] = VERTEX3D(VECTOR(-50.0f, 0.0f, 50.0f),  VECTOR(-f45DegAng, 0.0f, f45DegAng),  0.0f, 0.0f);
	pyramid->vertices[2] = VERTEX3D(VECTOR(50.0f, 0.0f, -50.0f),  VECTOR(f45DegAng, 0.0f, -f45DegAng),  0.0f, 0.0f);
	pyramid->vertices[3] = VERTEX3D(VECTOR(50.0f, 0.0f, 50.0f),   VECTOR(f45DegAng, 0.0f, f45DegAng),   0.0f, 0.0f);
	pyramid->vertices[4] = VERTEX3D(VECTOR(0.0f, 100.0f, 0.0f),   VECTOR(0.0f, 1.0f, 0.0f),			    0.0f, 0.0f);
	pyramid->vertices[5] = VERTEX3D(VECTOR(0.0f, -100.0f, 0.0f),  VECTOR(0.0f, -1.0f, 0.0f),			0.0f, 0.0f);

	WORD *indices = pyramid->faces;
	*indices++ = 4;		*indices++ = 0;		*indices++ = 2;
	*indices++ = 4;		*indices++ = 1;		*indices++ = 0;
	*indices++ = 4;		*indices++ = 3;		*indices++ = 1;
	*indices++ = 4;		*indices++ = 2;		*indices++ = 3;
	*indices++ = 5;		*indices++ = 2;		*indices++ = 0;
	*indices++ = 5;		*indices++ = 0;		*indices++ = 1;
	*indices++ = 5;		*indices++ = 1;		*indices++ = 3;
	*indices++ = 5;		*indices++ = 3;		*indices++ = 2;

	ComputeObjectNormals(pyramid);

	plane = new OBJECT3D(4, 2);

	//  0*---------*1
	//   |		   |
	//   |		   |
	//	 |		   |
	//	 |		   |
	//   |		   |
	//	2*---------*3
	plane->vertices[0] = VERTEX3D(VECTOR(-200.0f, -100.0f, 200.0f),		VECTOR(-0.577f, 0.577f, 0.577f),	0.0f, 0.0f);
	plane->vertices[1] = VERTEX3D(VECTOR(200.0f, -100.0f, 200.0f),		VECTOR(0.577f, 0.577f, 0.577f),		0.0f, 0.0f);
	plane->vertices[2] = VERTEX3D(VECTOR(-200.0f, -100.0f, -200.0f),	VECTOR(-0.577f, 0.577f, -0.577f),	0.0f, 0.0f);
	plane->vertices[3] = VERTEX3D(VECTOR(200.0f, -100.0f, -200.0f),		VECTOR(0.577f, 0.577f, -0.577f),	0.0f, 0.0f);

	indices = plane->faces;
	*indices++ = 0;		*indices++ = 3;		*indices++ = 1;
	*indices++ = 0;		*indices++ = 2;		*indices++ = 3;

	ComputeObjectNormals(plane);

	MatrixIdentity(objTrans);
	MatrixIdentity(newrot);

	return TRUE;
}

void MoveObjects()
{
	MATRIX mat;
	
	if(left_drag) {
		MatrixMultiply(mat, objTrans, newrot);
		d3dDevice->SetWorldMatrix(mat);
	}
	else {
		SetMatrixRotateY(mat, fAng);
		MatrixMultiply(objTrans, objTrans, mat);
		d3dDevice->SetWorldMatrix(objTrans);
	}
}

void RenderObjects()
{
	d3dDevice->BeginRender();

	d3dDevice->Render(pyramid);
	d3dDevice->Render(plane);
	
	d3dDevice->DrawFrame();
}

void DestroyObjects()
{
	d3dDevice->Destroy();

	SAFE_DELETE(lightDir);
	SAFE_DELETE(lightAmb);
	SAFE_DELETE(m3dMat);

	SAFE_DELETE(pyramid);
	SAFE_DELETE(plane);
}