/* Copyright (C) 2000 Ross Diankov

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

// Name: device3d.cpp
// Purpose: Provides a class that takes care of all 3d rendering
//			This class stores the transformation matrices that define
//			a camera, a world, and a projection matrix. It also holds a 
//			pointer to the texture it will use for texture mapping. Basically
//			everything that has to do with the 3d rendering process.

#include <windows.h>
#include <math.h>
#include <stdio.h>

#include "math3d.h"
#include "object3d.h"
#include "device3d.h"

#define SAFE_DELETE(x) 	if(x) { delete (x);	(x) = NULL; } 
#define MSG(x)		MessageBox(NULL, x, "Application Message", MB_OK)
#define PUTPIXEL(dib, nr, ng, nb)	*dib = nb;		*(dib+1) = ng;		*(dib+2) = nr;


// Device3D constructor sets all values to default
Device3D::Device3D()
{	
	hWnd = NULL;
	isInit = false;

	shade_mode = SHADE_FLAT;
	render_mode = RENDER_FILLFACE;
	light_types = 0;

	// light types
	l3dAmbient = 0;
	l3dDir = 0;
	l3dPoint = 0;
	l3dSpot = 0;

	ZeroMemory(&kAmb, sizeof(COLOR3D));
	ZeroMemory(&kDifDir, sizeof(COLOR3D));
	ZeroMemory(&kDifPoint, sizeof(COLOR3D));
	ZeroMemory(&kDifSpot, sizeof(COLOR3D));
	
	// set up default white material
	mat = 0;
	tex = 0;

	ZeroMemory(&vCamPos, sizeof(VECTOR));

	MatrixIdentity(invworld);
	MatrixIdentity(world);
	MatrixIdentity(view);

	// default
	fNearPlane = 1.0f;
	fFarPlane = 1000.0f;
	SetProjectionMatrix(proj, G_PI * 0.25f, G_PI * 0.25f, 1.0f, 1.0f, 1000.0f);

	zbuffer = 0;
	surface = 0;
	hSurface = NULL;

	dwWidth = 0;
	dwHeight = 0;
	clipWidth = 0;
	clipHeight = 0;
}

// destructor checks if the app is deleted before 
// Destroy() is called
Device3D::~Device3D()
{
	if(!isInit) Destroy();
}

// Init is called to initiate the device and set it to the current window
bool Device3D::Init(HWND hwnd)
{
	if(isInit) return true;

	BITMAPINFOHEADER bi;
	HDC hdc;

	isInit = true;			// set the flag
	hWnd = hwnd;			// set the handle

	GetClientRect(hWnd, &rSrc);			// get the dimensions of the window
	dwWidth = rSrc.right - rSrc.left;
	dwHeight = rSrc.bottom - rSrc.top;

	// for some reason screws up when width and height are not multiples of 4
	dwWidth = dwWidth - (dwWidth % 4);
	dwHeight = dwHeight - (dwHeight % 4);

	rSrc.right = rSrc.left + dwWidth;
	rSrc.bottom = rSrc.top + dwHeight;

	hdc = GetDC(hWnd);

	dwBPP = 3;				// 3 bytes means 24bit color

	bi.biSize = sizeof(BITMAPINFOHEADER);
	bi.biWidth = dwWidth;
	bi.biHeight = dwHeight;
	bi.biPlanes = 1;
	bi.biBitCount = dwBPP << 3;		// convert to bits
	bi.biCompression = BI_RGB;
	bi.biSizeImage = 0;
	bi.biClrImportant = 0;
	bi.biClrUsed = 0;
	bi.biXPelsPerMeter = 0;
	bi.biYPelsPerMeter = 0;

	hSurface = CreateDIBSection(hdc, (LPBITMAPINFO)&bi, DIB_RGB_COLORS, (void**)&surface, NULL, 0);

	if(!hSurface) {
		MSG("Call to create bitmap failed");
		return false;
	}

	if(bi.biHeight > 0) {	// upside down surface where the origin is at the lower right corner
		mem = surface + (dwWidth * dwBPP * (dwHeight-1));
		dwPitch = -dwWidth * (int)dwBPP;
	}
	else {					// right side up where the corner is the upper left
		mem = surface;
		dwPitch = dwWidth * (int)dwBPP;
	}

	// clear the screen and set the size of the screen in bytes
	srcMemory = dwHeight * dwWidth * dwBPP;
	memset(surface, 0xFF, srcMemory);

	// release the dc
	ReleaseDC(hWnd, hdc);

	// create a zbuffer
	zbuffer = new SCALAR[dwWidth * dwHeight];

	clipWidth = dwWidth - 1;					// for clipping since the screen goes from 0 to dwWidth-1
	clipHeight = dwHeight - 1;

	return true;
}

// Destroy() should be called every time the application closes or switches windows
void Device3D::Destroy()
{
	isInit = false;

	hWnd = NULL;

	// release the offscreen surface memory
	DeleteObject(hSurface);
	hSurface = NULL;

	// get rid of the pointers
	surface = 0;
	mem = 0;

	// deallocate the zbuffer memory
	delete[] zbuffer;

	l3dAmbient = 0;
	l3dDir = 0;
	l3dPoint = 0;
	l3dSpot = 0;
	light_types = 0;

	// set pointer to 0 but don't delete
	tex = 0;
	mat = 0;
}

void Device3D::ShowStats(HDC& hdc)
{
	static SCALAR fFPS = 0.0f;
	static DWORD dwFrames = 0;
	static DWORD dwLastTime = 0;
	char str[80];
	DWORD dwTime;
	
	dwFrames++;

	dwTime = timeGetTime();
	if(dwTime - dwLastTime > 1000) {
		fFPS = dwFrames / (SCALAR)(dwTime - dwLastTime) * 1000.0f;
		dwLastTime = timeGetTime();
		dwFrames = 0;
	}

	sprintf(str, "%dx%d fps: %3.2f", dwWidth, dwHeight, fFPS);

	SetTextColor(hdc, RGB(255, 255, 255));
	SetBkMode(hdc, TRANSPARENT);
	TextOut(hdc, 10, 10, str, strlen(str));
}

// copy the offscreen buffer to the screen
void Device3D::DrawFrame()
{
	HDC hdc = GetDC(hWnd);
	HDC memdc = CreateCompatibleDC(hdc);

	HBITMAP hOld = (HBITMAP)SelectObject(memdc, hSurface);
	
	ShowStats(memdc);

	BitBlt(hdc, rSrc.left, rSrc.top, dwWidth, dwHeight, memdc, 0, 0, SRCCOPY);

	SelectObject(memdc, hOld);

	DeleteDC(memdc);
	ReleaseDC(hWnd, hdc);
}

// should be called every time a new frame begins to 
// clear the zbuffer and offscreen memory
void Device3D::BeginRender()
{
	memset(surface, 0, srcMemory);						// clear the screen to white
	memset(zbuffer, 0, dwWidth*dwHeight*sizeof(SCALAR));	// clear the zbuffer
}

// the device already holds a poitner to the window
// so all we have to do is query the surface rect from the window handle 
bool Device3D::Update()
{
	RECT rc;

	GetWindowRect(hWnd, &rc);				// get the window dimensions
	
	// if the dimensions are different
	if( (rc.right - rc.left) != dwWidth ||
		(rc.bottom - rc.top) != dwHeight) {
		// deallocate the buffers and make new ones with the dims
		DeleteObject(hSurface);
		hSurface = NULL;
		mem = 0;
		surface = 0;		
		delete[] zbuffer;
		
		isInit = FALSE;

		// since all Init() does is initiate the buffers just call it with our window
		return Init(hWnd);
	}

	return true;
}

// SetSurfaceRect sets the destination surface of where the device
// will be drawing to a custom rectangle
bool Device3D::SetSurfaceRect(RECT& rc)
{
	rSrc = rc;

	if( (rSrc.right - rSrc.left) != dwWidth ||
		(rSrc.bottom - rSrc.top) != dwHeight) {

		BITMAPINFOHEADER bi;
		HDC hdc;

		// deallocate the buffers and make new ones with the dims
		DeleteObject(hSurface);
		hSurface = NULL;
		mem = 0;
		surface = 0;

		dwWidth = rSrc.right - rSrc.left;
		dwHeight = rSrc.bottom - rSrc.top;

		dwWidth = dwWidth - (dwWidth % 4);
		dwHeight = dwHeight - (dwHeight % 4);

		rSrc.right = rSrc.left + dwWidth;
		rSrc.bottom = rSrc.top + dwHeight;

		hdc = GetDC(hWnd);

		dwBPP = 3;				// 3 bytes means 24bit color

		bi.biSize = sizeof(BITMAPINFOHEADER);
		bi.biWidth = dwWidth;
		bi.biHeight = dwHeight;
		bi.biPlanes = 1;
		bi.biBitCount = dwBPP << 3;		// convert to bits
		bi.biCompression = BI_RGB;
		bi.biSizeImage = 0;
		bi.biClrImportant = 0;
		bi.biClrUsed = 0;
		bi.biXPelsPerMeter = 0;
		bi.biYPelsPerMeter = 0;

		hSurface = CreateDIBSection(hdc, (LPBITMAPINFO)&bi, DIB_RGB_COLORS, (void**)&surface, NULL, 0);

		if(!hSurface) {
			MSG("Call to create bitmap failed");
			return false;
		}

		if(bi.biHeight > 0) {	// upside down surface where the origin is at the lower right corner
			mem = surface + (dwWidth * dwBPP * (dwHeight-1));
			dwPitch = -dwWidth * dwBPP;
		}
		else {					// right side up where the corner is the upper left
			mem = surface;
			dwPitch = dwWidth * dwBPP;
		}

		// clear the screen and set the size of the screen in bytes
		srcMemory = dwHeight * dwWidth * dwBPP;
		memset(surface, 0xFF, srcMemory);

		// release the dc
		ReleaseDC(hWnd, hdc);

		// create a zbuffer
		zbuffer = new SCALAR[dwWidth * dwHeight];

		clipWidth = dwWidth - 1;					// for clipping since the screen goes from 0 to dwWidth-1
		clipHeight = dwHeight - 1;
	}

	return true;
}

//////////////////////////////////////////////////////////////////////////////////////
// The 3d pipeline will do all its color computing and light transforming in		//
// model space because of the simplicity to transform a few vectors instead of		//
// hundreds of normals. Things that need to be recomputed every time a new light	//
// is added or removed or one of the transformation matrices are changed.			//
//////////////////////////////////////////////////////////////////////////////////////

void Device3D::SetWorldMatrix(MATRIX& w)
{
	world = w;

	// get the inverse of the world matrix notice how an orthogonal matrix's inverse is its transpose
	invworld._11 = world._11;	invworld._12 = world._21;	invworld._13 = world._31;	invworld._14 = 0.0f;
	invworld._21 = world._12;	invworld._22 = world._22;	invworld._23 = world._32;	invworld._24 = 0.0f;
	invworld._31 = world._13;	invworld._32 = world._23;	invworld._33 = world._33;	invworld._34 = 0.0f;
	invworld._41 = -world._41;	invworld._42 = -world._42;	invworld._43 = -world._43;	invworld._44 = 1.0f;

	// get world coordinates of cam position
	vCamPos.x = - view._41 * view._11 - view._42 * view._12 - view._43 * view._13;
	vCamPos.y = - view._41 * view._21 - view._42 * view._22 - view._43 * view._23;
	vCamPos.z = - view._41 * view._31 - view._42 * view._32 - view._43 * view._33;

	VectorMatrixMultiply(vCamPos, invworld, vCamPos);
	
	// we don't need to do ambient because ambient doesn't have position or direction
	if(light_types & LIGHT_DIRECTIONAL) {
		vtDir.x = invworld._11 * l3dDir->dir.x + invworld._21 * l3dDir->dir.y + invworld._31 * l3dDir->dir.z;
		vtDir.y = invworld._12 * l3dDir->dir.x + invworld._22 * l3dDir->dir.y + invworld._32 * l3dDir->dir.z;
		vtDir.z = invworld._13 * l3dDir->dir.x + invworld._23 * l3dDir->dir.y + invworld._33 * l3dDir->dir.z;
	}
	if(light_types & LIGHT_POINT) VectorMatrixMultiply(vtPoint, invworld, l3dPoint->pos);
	if(light_types & LIGHT_SPOT) {
		VectorMatrixMultiply(vtSpotPoint, invworld, l3dSpot->pos);
		vtSpotDir.x = invworld._11 * l3dSpot->dir.x + invworld._21 * l3dSpot->dir.y + invworld._31 * l3dSpot->dir.z;
		vtSpotDir.y = invworld._12 * l3dSpot->dir.x + invworld._22 * l3dSpot->dir.y + invworld._32 * l3dSpot->dir.z;
		vtSpotDir.z = invworld._13 * l3dSpot->dir.x + invworld._23 * l3dSpot->dir.y + invworld._33 * l3dSpot->dir.z;
	}
}

// sets the camera matrix that'll be used to transform points to camera space. 
// vPos is the position of the camera in world coordinates
// vLookAt is a point in world coordinates that the camera is looking at
// vUp is the up vector of the camera
// Copyright by Microsoft DirectX 6.1 from 'd3dutil.cpp'
void Device3D::SetViewMatrix(VECTOR& vFrom, VECTOR& vAt, VECTOR& vWorldUp)
{
	// get the z basis vector, which poitns straight ahead. This is the difference
	// from the eyepoint to the lookat point.
	VECTOR vView = vAt - vFrom;
	SCALAR fLength = Magnitude(vView);

	vView = vView / fLength;

	// get the dot product and calculate the projection of the 
	// z vector onto the y vector. Once calculated, multiply it
	// by the z vector to calculate the vector that the y vector
	// is displaced by from being perpendicular to it.
	VECTOR vUp = vWorldUp - vView * DotProduct(vWorldUp, vView);
	fLength = Magnitude(vUp);
	vUp = vUp / fLength;

	VECTOR vRight = CrossProduct(vUp, vView);

	// set the view matrix
	// all the view matrix is is the inverse of it's transformation
	// for orthogonal matrices, the inverse is it's transposed matrix.
	// Since our 3x3 upper matrix is orthogonal, just transpose it, then
	// inverse the translating values
	view._11 = vRight.x;	view._21 = vRight.y;	view._31 = vRight.z;
	view._12 = vUp.x;		view._22 = vUp.y;		view._32 = vUp.z;
	view._13 = vView.x;		view._23 = vView.y;		view._33 = vView.z;
	
	view._41 = - DotProduct(vFrom, vRight);
	view._42 = - DotProduct(vFrom, vUp);
	view._43 = - DotProduct(vFrom, vView);

	view._14 = 0.0f;
	view._24 = 0.0f;
	view._34 = 0.0f;
	view._44 = 1.0f;

	VectorMatrixMultiply(vCamPos, invworld, vFrom);
}

void Device3D::SetViewMatrix(MATRIX& v)
{
	view = v;

	// get world coordinates of cam position
	vCamPos.x = - view._41 * view._11 - view._42 * view._12 - view._43 * view._13;
	vCamPos.y = - view._41 * view._21 - view._42 * view._22 - view._43 * view._23;
	vCamPos.z = - view._41 * view._31 - view._42 * view._32 - view._43 * view._33;

	// and transform them to world coordinates
	VectorMatrixMultiply(vCamPos, invworld, vCamPos);
}

void Device3D::SetProjectMatrix(MATRIX& p)
{
	proj = p;

	fNearPlane = - proj._43 / proj._33;
	fFarPlane = (fNearPlane * proj._33) / (fNearPlane - 1.0f);
}

// add a light to the scene
// or if light alrady exists of the same type, replace it
// A note to make on the lights is that the device acquires a poitner
// to the light and doesn't allocate any memory or delete them.
void Device3D::SetLight(DWORD lt, LIGHT3D* light)
{
	// check every light option and if that light is already supported
	// then replace the light
	if(lt & LIGHT_AMBIENT) {
		l3dAmbient = light;

		if(mat) {
			kAmb.r = mat->ambient.r * l3dAmbient->color.r;
			kAmb.g = mat->ambient.g * l3dAmbient->color.g;
			kAmb.b = mat->ambient.b * l3dAmbient->color.b;
		}
	}

	if(lt & LIGHT_DIRECTIONAL) {
		l3dDir = light;
		vtDir.x = invworld._11 * l3dDir->dir.x + invworld._21 * l3dDir->dir.y + invworld._31 * l3dDir->dir.z;
		vtDir.y = invworld._12 * l3dDir->dir.x + invworld._22 * l3dDir->dir.y + invworld._32 * l3dDir->dir.z;
		vtDir.z = invworld._13 * l3dDir->dir.x + invworld._23 * l3dDir->dir.y + invworld._33 * l3dDir->dir.z;

		if(mat) {
			kDifDir.r = mat->diffuse.r * l3dDir->color.r;
			kDifDir.g = mat->diffuse.g * l3dDir->color.g;
			kDifDir.b = mat->diffuse.b * l3dDir->color.b;
		}
	}
	if(lt & LIGHT_POINT) {
		l3dPoint = light;
		VectorMatrixMultiply(vtPoint, invworld, l3dPoint->pos);

		if(mat) {
			kDifPoint.r = mat->diffuse.r * l3dPoint->color.r;
			kDifPoint.g = mat->diffuse.g * l3dPoint->color.g;
			kDifPoint.b = mat->diffuse.b * l3dPoint->color.b;
		}
	}
	if(lt & LIGHT_SPOT) {
		l3dSpot = light;
		VectorMatrixMultiply(vtSpotPoint, invworld, l3dSpot->pos);
		vtSpotDir.x = invworld._11 * l3dSpot->dir.x + invworld._21 * l3dSpot->dir.y + invworld._31 * l3dSpot->dir.z;
		vtSpotDir.y = invworld._12 * l3dSpot->dir.x + invworld._22 * l3dSpot->dir.y + invworld._32 * l3dSpot->dir.z;
		vtSpotDir.z = invworld._13 * l3dSpot->dir.x + invworld._23 * l3dSpot->dir.y + invworld._33 * l3dSpot->dir.z;
	}
		
	light_types |= lt;
}

// gets a pointer to the light
void Device3D::GetLight(DWORD lt, LIGHT3D** light)
{
	if(lt & LIGHT_AMBIENT)		(*light) = l3dAmbient;
	if(lt & LIGHT_DIRECTIONAL)	(*light) = l3dDir;
	if(lt & LIGHT_POINT)		(*light) = l3dPoint;
	if(lt & LIGHT_SPOT)			(*light) = l3dSpot;
}

// since the device holds a pointer to the lights
// just reset the lights
void Device3D::RemoveLight(DWORD lt)
{
	if(lt & LIGHT_AMBIENT) {
		l3dAmbient = 0;
		ZeroMemory(&kAmb, sizeof(COLOR3D));
	}
	if(lt & LIGHT_DIRECTIONAL) {
		l3dDir = 0;
		ZeroMemory(&kDifDir, sizeof(COLOR3D));
	}
	if(lt & LIGHT_POINT) {
		l3dPoint = 0;
		ZeroMemory(&kDifPoint, sizeof(COLOR3D));
	}
	if(lt & LIGHT_SPOT) {
		l3dSpot = 0;
		ZeroMemory(&kDifSpot, sizeof(COLOR3D));
	}

	light_types &= ~lt;
}

void Device3D::SetMaterial(MATERIAL3D *m)
{
	mat = m;

	if(light_types & LIGHT_AMBIENT) {
		kAmb.r = mat->ambient.r * l3dAmbient->color.r;
		kAmb.g = mat->ambient.g * l3dAmbient->color.g;
		kAmb.b = mat->ambient.b * l3dAmbient->color.b;
	}
	if(light_types & LIGHT_DIRECTIONAL) {
		kDifDir.r = mat->diffuse.r * l3dDir->color.r;
		kDifDir.g = mat->diffuse.g * l3dDir->color.g;
		kDifDir.b = mat->diffuse.b * l3dDir->color.b;
	}
	if(light_types & LIGHT_POINT) {
		kDifPoint.r = mat->diffuse.r * l3dPoint->color.r;
		kDifPoint.g = mat->diffuse.g * l3dPoint->color.g;
		kDifPoint.b = mat->diffuse.b * l3dPoint->color.b;
	}
}

// How the clipping works is that all points are clipped to one
// plane then another until they are clipped to all planes. 
// This is called Sutherland-Hodgman Clipping.
// The TRANSFORMEDDATA struct is a struct that is filled with the new
// clipped points of the functions
void Device3D::Clip_Near(POLYGON* data, VERTEX3D& v1, VERTEX3D& v2)
{
	SCALAR m;								// delta to linearly interpolate to new point
	char v1in = 1, v2in = 1;

	if(v1.pos.z < 0.0f) v1in = 0;
	if(v2.pos.z < 0.0f) v2in = 0;

	// 3 cases of the two points 
	// both inside, a inside and b outside, a outside and b inside

	// both out
	if( !v1in && !v2in ) return;

	// both in
	if( (v1in && v2in) || v1.pos.z == v2.pos.z ) {
		data->p3d[data->numPoints++] = v1;
		return;
	}

	m = - v1.pos.z / (v2.pos.z - v1.pos.z);	

	// one is inside, or one is outside (whichever u want, it's all the same)
	// if it's entering   (v1 is out and v2 is in), then add in point 2
	if( v1in ) data->p3d[data->numPoints++] = v1;

	data->p3d[data->numPoints].pos.x = v1.pos.x + m * (v2.pos.x - v1.pos.x);
	data->p3d[data->numPoints].pos.y = v1.pos.y + m * (v2.pos.y - v1.pos.y);
	data->p3d[data->numPoints].pos.z = 0.0f;
	if(shade_mode & SHADE_GOURAUD) 
		data->p3d[data->numPoints].color = v1.color + m * (v2.color - v1.color);
	data->numPoints++;
}

void Device3D::Clip_Left(POLYGON* data, VERTEX3D& v1, VERTEX3D& v2)
{
	SCALAR m;							// delta to linearly interpolate to new point
	char v1in = 1, v2in = 1;

	if(v1.pos.x < -v1.pos.z) v1in = 0;
	if(v2.pos.x < -v2.pos.z) v2in = 0;

	// 3 cases of the two points 
	// both inside, a inside and b outside, a outside and b inside

	// both out
	if( !v1in && !v2in) return;

	// both in
	if(v1in && v2in) {
		data->p3d[data->numPoints++] = v1;
		return;
	}

	m =  (v1.pos.x + v1.pos.z ) / ( -v2.pos.z + v1.pos.z - v2.pos.x + v1.pos.x);

	// one is inside, or one is outside (whichever u want, it's all the same)
	// if it's entering   (v1 is out and v2 is in), then add in point 2
	if( v1in ) data->p3d[data->numPoints++] = v1;

	data->p3d[data->numPoints].pos.x = v1.pos.x + m * (v2.pos.x - v1.pos.x);
	data->p3d[data->numPoints].pos.y = v1.pos.y + m * (v2.pos.y - v1.pos.y);
	data->p3d[data->numPoints].pos.z = -data->p3d[data->numPoints].pos.x;
	if(shade_mode & SHADE_GOURAUD) 
		data->p3d[data->numPoints].color = v1.color + m * (v2.color - v1.color);
	data->numPoints++;
}

void Device3D::Clip_Right(POLYGON* data, VERTEX3D& v1, VERTEX3D& v2)
{
	SCALAR m;							// delta to linearly interpolate to new point
	char v1in = 1, v2in = 1;

	if(v1.pos.x > v1.pos.z) v1in = 0;
	if(v2.pos.x > v2.pos.z) v2in = 0;

	// 3 cases of the two points 
	// both inside, a inside and b outside, a outside and b inside

	// both out
	if( !v1in && !v2in) return;

	// both in
	if(v1in && v2in) {
		data->p3d[data->numPoints++] = v1;
		return;
	}

	m = (v1.pos.x - v1.pos.z) / ( - v2.pos.x + v1.pos.x + v2.pos.z - v1.pos.z);		

	// one is inside, or one is outside (whichever u want, it's all the same)
	// if it's entering   (v1 is out and v2 is in), then add in point 2
	if( v1in ) data->p3d[data->numPoints++] = v1;

	data->p3d[data->numPoints].pos.x = v1.pos.x + m * (v2.pos.x - v1.pos.x);
	data->p3d[data->numPoints].pos.y = v1.pos.y + m * (v2.pos.y - v1.pos.y);
	data->p3d[data->numPoints].pos.z = data->p3d[data->numPoints].pos.x;
	if(shade_mode & SHADE_GOURAUD) 
		data->p3d[data->numPoints].color = v1.color + m * (v2.color - v1.color);
	data->numPoints++;
}

void Device3D::Clip_Top(POLYGON* data, VERTEX3D& v1, VERTEX3D& v2)
{
	SCALAR m;							// delta to linearly interpolate to new point
	char v1in = 1, v2in = 1;

	if(v1.pos.y > v1.pos.z) v1in = 0;
	if(v2.pos.y > v2.pos.z) v2in = 0;

	// 3 cases of the two points 
	// both inside, a inside and b outside, a outside and b inside

	// both out
	if( !v1in && !v2in) return;

	// both in
	if(v1in && v2in) {
		data->p3d[data->numPoints++] = v1;
		return;
	}

	m = (v1.pos.y - v1.pos.z) / ( - v2.pos.y + v1.pos.y + v2.pos.z - v1.pos.z);

	// one is inside, or one is outside (whichever u want, it's all the same)
	// if it's entering   (v1 is out and v2 is in), then add in point 2
	if( v1in ) data->p3d[data->numPoints++] = v1;
	
	data->p3d[data->numPoints].pos.x = v1.pos.x + m * (v2.pos.x - v1.pos.x);
	data->p3d[data->numPoints].pos.y = v1.pos.y + m * (v2.pos.y - v1.pos.y);
	data->p3d[data->numPoints].pos.z = data->p3d[data->numPoints].pos.y;
	if(shade_mode & SHADE_GOURAUD) 
		data->p3d[data->numPoints].color = v1.color + m * (v2.color - v1.color);
	data->numPoints++;
}

void Device3D::Clip_Bottom(POLYGON* data, VERTEX3D& v1, VERTEX3D& v2)
{
	SCALAR m;							// delta to linearly interpolate to new point
	char v1in = 1, v2in = 1;

	if(v1.pos.y < -v1.pos.z) v1in = 0;
	if(v2.pos.y < -v2.pos.z) v2in = 0;

	// 3 cases of the two points 
	// both inside, a inside and b outside, a outside and b inside

	// both out
	if( !v1in && !v2in) return;

	// both in
	if(v1in && v2in) {
		data->p3d[data->numPoints++] = v1;
		return;
	}

	m = (v1.pos.y + v1.pos.z) / ( - v2.pos.z + v1.pos.z - v2.pos.y + v1.pos.y);
	
	// one is inside, or one is outside (whichever u want, it's all the same)
	// if it's entering   (v1 is out and v2 is in), then add in point 2
	if( v1in ) data->p3d[data->numPoints++] = v1;

	data->p3d[data->numPoints].pos.x = v1.pos.x + m * (v2.pos.x - v1.pos.x);
	data->p3d[data->numPoints].pos.y = v1.pos.y + m * (v2.pos.y - v1.pos.y);
	data->p3d[data->numPoints].pos.z = v1.pos.z + m * (v2.pos.z - v1.pos.z);
	if(shade_mode & SHADE_GOURAUD) 
		data->p3d[data->numPoints].color = v1.color + m * (v2.color - v1.color);
	data->numPoints++;
}

// Computes the screen coordinates of all the vertices of the object
// with a few additional stuff like z value and vertex color. Also clips
// in 3D a point to the viewing frustrum.
TRANSFORMEDDATA* Device3D::ComputeScreenCoords(OBJECT3D* obj)
{
	int i, j;
	TRANSFORMEDDATA* data;			// the 2D data that defines the object
	WORD faceIndices[MAX_FACES];	// the indices of the faces showing
	WORD face_norm[MAX_NORMALS];	// the normals of the faces showing	
	VERTEX3D *pts;					// stores transformed screen points
	VECTOR vTemp;
	MATRIX trans;					// transforms a vector in world to projection space

	// the matrix that transforms coordinates to projection coordinates is
	// TotalTransformation = World * View * Projection
	MatrixMultiply(trans, world, view);	

	pts = new VERTEX3D[obj->numVertices];
	data = new TRANSFORMEDDATA;

	WORD* dindices = faceIndices;
	WORD* objindices = obj->faces;	
	
	// back face culling
	for(i = 0; i < obj->numFaces; i++) {
		// if it's greater than 0, then face is facing camera
		// also if all points share an outside plane, then they aren't in the view frustrum
		if( (DotProduct((VECTOR)(vCamPos - obj->vertices[*objindices].pos), obj->face_normals[i]) >= 0.0f) ) {

			// increment the number of faces actually showing
			face_norm[data->numFaces++] = i;
			*dindices++ = *(objindices  );		// copy the indices of the object to the data structure
			*dindices++ = *(objindices+1);
			*dindices++ = *(objindices+2);
		}
		
		objindices += 3;						// don't forget!!!
	}
	
	// project and light
	i = obj->numVertices - 1;
	do {
		SCALAR objcol;

		pts[i].pos.z =	obj->vertices[i].pos.x * trans._13 + obj->vertices[i].pos.y * trans._23 + 
						obj->vertices[i].pos.z * trans._33 + trans._43;
		
		// if any vertex is outside of far clipping plane, then just don't render the object
		if(pts[i].pos.z > fFarPlane) return data;

		pts[i].pos.x =	obj->vertices[i].pos.x * trans._11 + obj->vertices[i].pos.y * trans._21 + 
						obj->vertices[i].pos.z * trans._31 + trans._41;
		pts[i].pos.y =	obj->vertices[i].pos.x * trans._12 + obj->vertices[i].pos.y * trans._22 + 
						obj->vertices[i].pos.z * trans._32 + trans._42;

		objcol = 0.0f;

		if(shade_mode & SHADE_GOURAUD) {					// vertex color will be used

			if(light_types & LIGHT_DIRECTIONAL) objcol += max(DotProduct(vtDir, obj->vertices[i].normal), 0.0f);	
		}

		pts[i].color = objcol;

		i--;
	} while( i >= 0 );

	data->faces = new POLYGON[data->numFaces];

	dindices = faceIndices;

	// now we have a data structure ready for clipping with everything set
	i = data->numFaces - 1;
	do {
		WORD next;
		POLYGON face1, face2;

		face1.p3d = new VERTEX3D[10];
		face2.p3d = new VERTEX3D[10];

		face1.numPoints = 3;
		face1.p3d[0] = pts[faceIndices[i*3]];
		face1.p3d[1] = pts[faceIndices[i*3+1]];
		face1.p3d[2] = pts[faceIndices[i*3+2]];

		// clip to all planes								
		for(j = 0; j < face1.numPoints; j++) {			
			next = j + 1;
			if(next == face1.numPoints) next = 0;
			Clip_Near(&face2, face1.p3d[j], face1.p3d[next]);
		}
		face1.numPoints = 0;	
		for(j = 0; j < face2.numPoints; j++) {	
			next = j + 1;
			if(next == face2.numPoints) next = 0;
			Clip_Left(&face1, face2.p3d[j], face2.p3d[next]);
		}
		face2.numPoints = 0;
		for(j = 0; j < face1.numPoints; j++) {				
			next = j + 1;
			if(next == face1.numPoints) next = 0;
			Clip_Right(&face2, face1.p3d[j], face1.p3d[next]);
		}
		face1.numPoints = 0;								
		for(j = 0; j < face2.numPoints; j++) {			
			next = j + 1;
			if(next == face2.numPoints) next = 0;
			Clip_Top(&face1, face2.p3d[j], face2.p3d[next]);
		}
		face2.numPoints = 0;							
		for(j = 0; j < face1.numPoints; j++) {
			next = j + 1;
			if(next == face1.numPoints) next = 0;
			Clip_Bottom(&face2, face1.p3d[j], face1.p3d[next]);
		}	

		data->faces[i].p2d = 0;

		if(face2.numPoints > 0) {
	
			data->faces[i].numPoints = face2.numPoints;
			data->faces[i].p2d = new POINT2D[face2.numPoints];

			j = face2.numPoints - 1;
			do {
				data->faces[i].p2d[j].z = face2.p3d[j].pos.z;
				VectorMatrixMultiply(face2.p3d[j].pos, proj, face2.p3d[j].pos);
				data->faces[i].p2d[j].x = (int)((1.0f + face2.p3d[j].pos.x) * 0.5f * clipWidth);	
				data->faces[i].p2d[j].y = (int)((1.0f - face2.p3d[j].pos.y) * 0.5f * clipHeight);
				data->faces[i].p2d[j].color = face2.p3d[j].color;

				j--;
			} while(j >= 0);
		
			if(shade_mode & SHADE_FLAT) data->faces[i].norm = obj->face_normals[face_norm[i]];
		}

		delete[] face1.p3d;		
		delete[] face2.p3d;

		i--;
	} while( i >= 0 );

	delete[] pts;

	return data;							// return a pointer to the data and that's it!! 
}

// Render an object to the screen. Compute the screen coords and 
// pass them to a separate function that renders them to the
// mode the device is set to
void Device3D::Render(OBJECT3D* obj)
{
	if(render_mode & RENDER_FILLFACE) {

		if(!obj->numFaces) return;

		TRANSFORMEDDATA* data;				// a pointer to an array of the screen coords for the object

		data = ComputeScreenCoords(obj);

		if(data->faces) {
			if(shade_mode & SHADE_FLAT)		 RenderFlat(data);
			if(shade_mode & SHADE_GOURAUD)   RenderGouraud(data);
			if(shade_mode & SHADE_PHONG)	 RenderGouraud(data);
			if(shade_mode & SHADE_WIREFRAME) RenderWireframe(data);
		}

		for(int i = 0; i < data->numFaces; i++ ) SAFE_DELETE(data->faces[i].p2d);
		SAFE_DELETE(data);
		
		return;
	}

	if(render_mode & RENDER_LINESTRIP && !obj->numFaces) RenderLineStrip(obj);
	if(render_mode & RENDER_LINELIST) {

	}
	if(render_mode & RENDER_POINT) {

	}
}

// object has to face data
void Device3D::RenderLineStrip(OBJECT3D* obj)
{
	int i;
	MATRIX trans;
	POLYGON face1, face;
	int top, bottom;
	LPBYTE screen;
	
	face1.p3d = new VERTEX3D[obj->numVertices * 2];
	face.p3d = new VERTEX3D[obj->numVertices * 2];

	MatrixMultiply(trans, world, view);

	face1.numPoints = obj->numVertices;
	for(i = 0; i < obj->numVertices; i++) VectorMatrixMultiply(face1.p3d[i].pos, trans, obj->vertices[i].pos);

	// clip to all planes
	// since we're clipping a line and not a closed polygon, we'll have to leave the last edge
	// but another problem arises because if the last edge isn't checked, then the last point
	// (which can be clipped) doesn't show up in the array
	for(i = 0; i < face1.numPoints-1; i++) Clip_Near(&face, face1.p3d[i], face1.p3d[i+1]);
	Clip_Near(&face, face1.p3d[face1.numPoints-1], face1.p3d[0]);
	if(face.p3d[face.numPoints-1].pos != face1.p3d[face1.numPoints-1].pos) face.numPoints--;

	face1.numPoints = 0;	
	for(i = 0; i < face.numPoints-1; i++) Clip_Left(&face1, face.p3d[i], face.p3d[i+1]);
	Clip_Left(&face1, face.p3d[face.numPoints-1], face.p3d[0]);
	if(face1.p3d[face1.numPoints-1].pos != face.p3d[face.numPoints-1].pos) face1.numPoints--;

	face.numPoints = 0;
	for(i = 0; i < face1.numPoints-1; i++) Clip_Right(&face, face1.p3d[i], face1.p3d[i+1]);
	Clip_Right(&face, face1.p3d[face1.numPoints-1], face1.p3d[0]);
	if(face.p3d[face.numPoints-1].pos != face1.p3d[face1.numPoints-1].pos) face.numPoints--;

	face1.numPoints = 0;								
	for(i = 0; i < face.numPoints-1; i++) Clip_Top(&face1, face.p3d[i], face.p3d[i+1]);
	Clip_Top(&face1, face.p3d[face.numPoints-1], face.p3d[0]);
	if(face1.p3d[face1.numPoints-1].pos != face.p3d[face.numPoints-1].pos) face1.numPoints--;

	face.numPoints = 0;							
	for(i = 0; i < face1.numPoints-1; i++) Clip_Bottom(&face, face1.p3d[i], face1.p3d[i+1]);
	Clip_Bottom(&face, face1.p3d[face1.numPoints-1], face1.p3d[0]);
	if(face.p3d[face.numPoints-1].pos != face1.p3d[face1.numPoints-1].pos) face.numPoints--;

	delete[] face1.p3d;

	if(face.numPoints < 2) {
		delete[] face.p2d;
		return;
	}

	face1.p2d = new POINT2D[face.numPoints];
	i = face.numPoints - 1;
	do {
		VectorMatrixMultiply(face.p3d[i].pos, proj, face.p3d[i].pos);
		face1.p2d[i].x = (int)((1.0f + face.p3d[i].pos.x) * 0.5f * clipWidth);	
		face1.p2d[i].y = (int)((1.0f - face.p3d[i].pos.y) * 0.5f * clipHeight);
		i--;
	} while(i >= 0);

	delete[] face.p3d;

	for(i = 0; i < face.numPoints-1; i++) {

		top = i;
		bottom = i+1;
		if(face1.p2d[i].y > face1.p2d[i+1].y) {
			top = i+1;
			bottom = i;
		}

		screen = mem + (dwPitch * face1.p2d[top].y) + (dwBPP * face1.p2d[top].x);
		FillWireframeEdge(&screen, face1.p2d[top], face1.p2d[bottom]);
	}

	delete[] face1.p2d;
}

void Device3D::RenderWireframe(TRANSFORMEDDATA* data)
{
	int i, j, next;							// temporary values
	int top, bottom;
	POLYGON* face;							// the current face
	LPBYTE screen;

	// NOTE: All faces are showing and already clipped
	i = data->numFaces - 1;
	do {
		if(data->faces[i].numPoints == 0) continue;

		face = &data->faces[i];

		top = 0; 
		bottom = 0;
		j = face->numPoints - 1;
		do {
			if(face->p2d[j].y < face->p2d[top].y) top = j;
			if(face->p2d[j].y > face->p2d[bottom].y) bottom = j;
			j--;
		} while( j >= 0 );
		
		// go through the left edge
		screen = mem + (dwPitch * face->p2d[top].y) + (dwBPP * face->p2d[top].x);

		// there can be two tops, in this case we want to point to the left one
		j = top;
		next = j + 1;
		if(next == face->numPoints) next = 0;

		do {
			FillWireframeEdge(&screen, face->p2d[j], face->p2d[next]);

			j = next;
			next = j + 1;
			if(next == face->numPoints) next = 0;
		} while( j != bottom );

		// go through the right edge
		screen = mem + (dwPitch * face->p2d[top].y) + (dwBPP * face->p2d[top].x);

		// there can be two tops, in this case we want the pointer to point to the right one
		j = top;
		next = j - 1;
		if( next < 0 ) next = face->numPoints - 1;

		do {
			FillWireframeEdge(&screen, face->p2d[j], face->p2d[next]);

			j = next;
			next = j - 1;
			if( next < 0 ) next = face->numPoints - 1;
		} while( j != bottom );

		i--;
	} while( i >= 0 );
}

void Device3D::RenderFlat(TRANSFORMEDDATA* data)
{
	int i, j, k;						// temporary values
	int top, bottom, next;				// the indices of the top and bottom vertices
	POLYGON* face;						// the current face
	SCANLINE *left, *right;				// the scanline data
	SCALAR zsx, zdx;
	SCALAR intensity; 
	BYTE nr, ng, nb;

	left = new SCANLINE[dwHeight];
	right = new SCANLINE[dwHeight];

	// NOTE: All faces are showing and already clipped
	i = data->numFaces - 1;
	do {
		if(data->faces[i].numPoints == 0) {
			i--;
			continue;
		}

		face = &data->faces[i];

		intensity = max(DotProduct(data->faces[i].norm, vtDir), 0.0f);
		nr = (BYTE)(min(kAmb.r + intensity * kDifDir.r, 1.0f) * 255.0f);
		ng = (BYTE)(min(kAmb.g + intensity * kDifDir.g, 1.0f) * 255.0f);
		nb = (BYTE)(min(kAmb.b + intensity * kDifDir.b, 1.0f) * 255.0f);

		// fill the top and bottom indices
		top = 0;
		bottom = 0;
		j = face->numPoints - 1;
		do {
			if(face->p2d[top].y > face->p2d[j].y) top = j;
			if(face->p2d[bottom].y < face->p2d[j].y) bottom = j;
			j--;
		} while( j >= 1 );
		
		// fill the left edge 
		// since the points are counter-clockwise, top + 1 will be in the left edge
		// until top + n is equal to bottom
		j = top;
		next = top + 1;
		if(next == face->numPoints) next = 0;

		do {
			FillFlatEdge(left, face->p2d[j], face->p2d[next]);

			j = next;
			next = j + 1;
			if(next == face->numPoints) next = 0;
		} while(j != bottom);

		j = bottom;
		next = bottom + 1;
		if(next == face->numPoints) next = 0;

		do {
			FillFlatEdge(right, face->p2d[next], face->p2d[j]);

			j = next;
			next = j + 1;
			if(next == face->numPoints) next = 0;
		} while(j != top);

		for(j = face->p2d[top].y; j < face->p2d[bottom].y; j++) {
			// pointer to the start of the current scanline in memory
			LPBYTE dib = mem + (dwPitch * j) + (left[j].x * dwBPP);
			
			zsx = left[j].z;
			zdx = (right[j].z - left[j].z) / (right[j].x - left[j].x);
		
			for(k = left[j].x; k < right[j].x; k++) {

				if(zbuffer[dwWidth*j + k] > zsx || zbuffer[dwWidth*j + k] == 0.0f ) {
					// since a DIBs color is defined by BGR
					PUTPIXEL(dib, nr, ng, nb);
					
					// put the new z value into the zbuffer
					zbuffer[dwWidth*j + k] = zsx;
				}
				
				dib += dwBPP;				// increment the pointer!!!
				zsx += zdx;
			}
		}

		i--;
	} while( i >= 0 );

	delete[] left;
	delete[] right;
}

void Device3D::RenderGouraud(TRANSFORMEDDATA* data)
{
	int i, j, k;						// temporary values
	int top, bottom, next;				// the indices of the top and bottom vertices
	POLYGON* face;						// the current face
	GSCANLINE *left, *right;			// the scanline data

	left = new GSCANLINE[dwHeight];
	right = new GSCANLINE[dwHeight];

	// NOTE: All faces are showing and already clipped
	i = data->numFaces - 1;
	do {
		if(data->faces[i].numPoints == 0) continue;

		face = &data->faces[i];
	
		// fill the top and bottom indices
		top = 0;
		bottom = 0;
		j = face->numPoints - 1;
		do {
			if(face->p2d[top].y > face->p2d[j].y) top = j;
			if(face->p2d[bottom].y < face->p2d[j].y) bottom = j;
			j--;
		} while( j >= 1 );
		
		// fill the left edge 
		// since the points are counter-clockwise, top + 1 will be in the left edge
		// until top + n is equal to bottom
		j = top;
		next = top + 1;
		if(next == face->numPoints) next = 0;

		do {
			FillGouraudEdge(left, face->p2d[j], face->p2d[next]);

			j = next;
			next = j + 1;
			if(next == face->numPoints) next = 0;
		} while(j != bottom);

		j = bottom;
		next = bottom + 1;
		if(next == face->numPoints) next = 0;

		do {
			FillGouraudEdge(right, face->p2d[next], face->p2d[j]);

			j = next;
			next = j + 1;
			if(next == face->numPoints) next = 0;
		} while(j != top);

		for(j = face->p2d[top].y; j < face->p2d[bottom].y; j++) {
			// pointer to the start of the current scanline in memory
			LPBYTE dib = mem + (dwPitch * j) + (left[j].x * dwBPP);

			SCALAR zsx, zdx;
			SCALAR gsx, gdx;
			SCALAR ir, ig, ib;		// the different intensities
			SCALAR temp = 1.0f / (right[j].x - left[j].x);

			zsx = left[j].z;
			zdx = (right[j].z - left[j].z) * temp;
			
			gsx = left[j].color;
			gdx = (right[j].color - left[j].color) * temp;

			for(k = left[j].x; k < right[j].x; k++) {

				if(zbuffer[dwWidth*j + k] > zsx || zbuffer[dwWidth*j + k] == 0.0f ) {

					ir = min(kAmb.r + gsx * kDifDir.r, 1.0f) * 255.0f;
					ig = min(kAmb.g + gsx * kDifDir.g, 1.0f) * 255.0f;
					ib = min(kAmb.b + gsx * kDifDir.b, 1.0f) * 255.0f;

					// since a DIBs color is defined by BGR
					PUTPIXEL(dib, (BYTE)ir, (BYTE)ig, (BYTE)ib);
					
					// put the new z value into the zbuffer
					zbuffer[dwWidth*j + k] = zsx;
				}
				
				dib += dwBPP;				// increment the pointer!!!
				zsx += zdx;
				gsx += gdx;
			}
		}

		i--;
	} while( i >= 0 );

	delete[] left;
	delete[] right;
}

void Device3D::RenderPhong(TRANSFORMEDDATA*)
{
}

////////////////////////////////////////////////////////////////////////////////////////
// Purpose is to fill an edge array with the x values of a line						  //
// defined from points (x0, y0) and (x1, y1) with Bresenham's Midpoint Line Algorithm //
// Then fills the scanline array with z-buffer values								  //
////////////////////////////////////////////////////////////////////////////////////////

// NOTE: screen is a pointer that points to the pointer of the screen memory.
// The reason for this is because we'll need to increment the screen pointer
// and use it for other calls of FillWireframeEdge.
void Device3D::FillWireframeEdge(LPBYTE* screen, POINT2D& a, POINT2D& b)
{
	int pixShift;							// how much to add to the memory address to be at the next pixel 
	BYTE nr, ng, nb;
	int y, dx, dy, incrE, incrNE, d;
	int incx, long_d, short_d;

	nr = (BYTE)(kDifDir.r * 255.0f);
	ng = (BYTE)(kDifDir.g * 255.0f);
	nb = (BYTE)(kDifDir.b * 255.0f);

	dx = b.x - a.x;
	dy = b.y - a.y;

	incx = 1;
	if(dx < 0) { 
		incx = -1; 
		dx = -dx; 
	}

	long_d = dy; 
	short_d = dx;
	if(dx > dy) { 
		long_d = dx; 
		short_d = dy; 
	}


	d = short_d * 2 - long_d;
	incrE = short_d * 2;
	incrNE = (short_d - long_d) * 2;

	pixShift = dwBPP * incx;

	PUTPIXEL(*screen, nr, ng, nb);	

	if(dx >= dy) {

		while(dx > 1) {

			d += incrE;
			if(d >= 0) {
				d += incrNE - incrE; 
				*screen += dwPitch;
			}

			dx--;
			*screen += pixShift;
			PUTPIXEL(*screen, nr, ng, nb);
		}

		// the edge can be flat
		if(a.y == b.y) {
			while(dx > 1) {
				dx--;
				*screen += pixShift;
				PUTPIXEL(*screen, nr, ng, nb);
			}
		}

		return;
	}

	y = a.y;
	while(y < b.y) {

		d += incrE;
		if(d >= 0) {
			d += incrNE - incrE; 
			*screen += pixShift;
		}

		y++;
		*screen += dwPitch;	
		PUTPIXEL(*screen, nr, ng, nb);
	}
}

void Device3D::FillFlatEdge(SCANLINE* edge, POINT2D& a, POINT2D& b)
{
	int x, y, dx, dy, incrE, incrNE, d;
	int incx, long_d, short_d;
	SCALAR zcurr, dz;				// zbuffer interpolation

	zcurr = a.z;

	dx = b.x - a.x;
	dy = b.y - a.y;
	x = a.x;
	y = a.y;

	incx = 1;
	if(dx < 0) { 
		incx = -1; 
		dx = -dx; 
	}

	long_d = dy; 
	short_d = dx;
	if(dx > dy) { 
		long_d = dx; 
		short_d = dy; 
	}


	d = short_d * 2 - long_d;
	incrE = short_d * 2;
	incrNE = (short_d - long_d) * 2;

	edge[y].x = x;
	edge[y].z = zcurr;

	if(dx >= dy) {
		dz = (b.z - a.z) / dx;

		while(y < b.y) {

			d += incrE;
			if(d >= 0) {
				y++; 
				d += incrNE - incrE; 
				edge[y].x = x;
				edge[y].z = zcurr;
			}

			x += incx;
			zcurr += dz;
		}

		edge[b.y].x = b.x;
		edge[b.y].z = b.z;

		return;
	}

	dz = (b.z - a.z) / dy;

	while(y < b.y) {

		d += incrE;
		if(d >= 0) {
			d += incrNE - incrE; 
			x += incx; 
		}

		y++;
		zcurr += dz;
		edge[y].x = x;
		edge[y].z = zcurr;
	}

	edge[b.y].x = b.x;
	edge[b.y].z = b.z;
}

void Device3D::FillGouraudEdge(GSCANLINE* edge, POINT2D& a, POINT2D& b)
{
	int x, y, dx, dy, incrE, incrNE, d;
	int incx, long_d, short_d;
	SCALAR zcurr, dz;				// zbuffer interpolation
	SCALAR gcurr, dg;				// color interpolation

	zcurr = a.z;
	gcurr = a.color;

	dx = b.x - a.x;
	dy = b.y - a.y;
	x = a.x;
	y = a.y;

	incx = 1;
	if(dx < 0) { 
		incx = -1; 
		dx = -dx; 
	}

	long_d = dy; 
	short_d = dx;
	if(dx > dy) { 
		long_d = dx; 
		short_d = dy; 
	}


	d = short_d * 2 - long_d;
	incrE = short_d * 2;
	incrNE = (short_d - long_d) * 2;

	edge[y].x = x;
	edge[y].z = zcurr;
	edge[y].color = gcurr;

	if(dx >= dy) {
		dz = (b.z - a.z) / dx;
		dg = (b.color - a.color) / dx;
		
		while(y < b.y) {

			d += incrE;
			if(d >= 0) {
				y++; 
				d += incrNE - incrE; 
				edge[y].x = x;
				edge[y].z = zcurr;
				edge[y].color = gcurr;
			}

			x += incx;
			zcurr += dz;
			gcurr += dg;
		}

		edge[b.y].x = b.x;
		edge[b.y].z = b.z;
		edge[b.y].color = b.color;

		return;
	}

	dz = (b.z - a.z) / dy;
	dg = (b.color - a.color) / dy;

	while(y < b.y) {

		d += incrE;
		if(d >= 0) {
			d += incrNE - incrE; 
			x += incx; 
		}

		y++;
		zcurr += dz;
		gcurr += dg;
		edge[y].x = x;
		edge[y].z = zcurr;
		edge[y].color = gcurr;
	}

	edge[b.y].x = b.x;
	edge[b.y].z = b.z;
	edge[b.y].color = b.color;
}

//////////////////////////
//	Helper Functions	//
//////////////////////////

// TestPointInObject tests if a point clicked on the screen actually
// clicks on a polygon. 
// dev		- the main device of the program that will provide the transformation matrices
// invproj	- the inverse projection matrix to transform from screen to view coordinates
// sx & sy	- the screen coordinates of the point testing
int TestPointInObject(Device3D* dev, OBJECT3D* obj, int sx, int sy)
{
	int i;
	SCALAR rx, ry, y;
	MATRIX view, world, proj, mat; 
	WORD ncross = 0, *indices;
	VECTOR *pts;

	dev->GetProjectMatrix(proj);
	dev->GetViewMatrix(view);
	dev->GetWorldMatrix(world);

	MatrixMultiply(mat, world, view);
	MatrixMultiply(mat, mat, proj);

	pts = new VECTOR[obj->numVertices];

	for(i = 0; i < obj->numVertices; i++) VectorMatrixMultiply(pts[i], mat, obj->vertices[i].pos);

	rx = ((SCALAR)sx * 2.0f / dev->clipWidth) - 1.0f;
	ry = 1.0f - ((SCALAR)sy * 2.0f / dev->clipHeight);

	indices = obj->faces;
	
	// algorithm goes through all the faces and checks the amount of times segments have
	// crossed the point's upper y-axis
	for(i = 0; i < obj->numFaces; i++) {
		
		if( (pts[*indices].x >= rx && pts[*(indices+1)].x < rx) || (pts[*indices].x <= rx && pts[*(indices+1)].x > rx) ) {
			// edge crosses point's yaxis
			y = pts[*indices].y + (rx - pts[*indices].x) * (pts[*(indices+1)].y - pts[*indices].y) / (pts[*(indices+1)].x - pts[*indices].x);

			if ( y > ry )	ncross++;      
		}    
		if( (pts[*(indices+1)].x >= rx && pts[*(indices+2)].x < rx) || (pts[*(indices+1)].x <= rx && pts[*(indices+2)].x > rx) ) {
			// edge crosses point's yaxis
			y = pts[*(indices+1)].y + (rx - pts[*(indices+1)].x) * (pts[*(indices+2)].y - pts[*(indices+1)].y) / (pts[*(indices+2)].x - pts[*(indices+1)].x);

			if ( y > ry )	ncross++;      
		}    
		if( (pts[*(indices+2)].x >= rx && pts[*indices].x < rx) || (pts[*(indices+2)].x <= rx && pts[*indices].x > rx) ) {
			// edge crosses point's yaxis
			y = pts[*(indices+2)].y + (rx - pts[*(indices+2)].x) * (pts[*indices].y - pts[*(indices+2)].y) / (pts[*indices].x - pts[*(indices+2)].x);

			if ( y > ry )	ncross++;      
		}    

		if (ncross%2) return 1;

		indices += 3;
	}

	return 0;
}