/* Copyright (C) 2000 Ross Diankov

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

// Name: device3d.h
// Purpose: Provides a class that takes care of all 3d rendering
//			This class stores the transformation matrices that define
//			a camera, a world, and a projection matrix. It also holds a 
//			pointer to the texture it will use for texture mapping. Basically
//			everything that has to do with the 3d rendering process.

#ifndef DEVICE3D_CLASS
#define DEVICE3D_CLASS

// render modes
#define RENDER_POINT		0x00000001
#define RENDER_LINESTRIP	0x00000002
#define RENDER_LINELIST		0x00000004
#define RENDER_FILLFACE		0x00000008

// shading modes
#define SHADE_WIREFRAME		0x00000001
#define	SHADE_FLAT			0x00000002
#define	SHADE_GOURAUD		0x00000004
#define	SHADE_PHONG			0x00000008

// light modes
#define LIGHT_AMBIENT		0x00000001
#define LIGHT_DIRECTIONAL	0x00000002
#define LIGHT_POINT			0x00000004
#define LIGHT_SPOT			0x00000008

#define MAX_FACES			200
#define MAX_NORMALS			100	


struct POINT2D
{
	int x, y;
	SCALAR z, color;
};

// helper objects
struct POLYGON
{
	POLYGON()  { numPoints = 0; }

	int numPoints;

	VECTOR norm;				// pointer to normal in 3d space

	union {
		POINT2D *p2d;
		VERTEX3D *p3d;
	};
};

// Defines a data structure that holds the transformed data of an object.
// So all that needs to be done is some kind of algorithm that draws
// n-gons onto the screen
struct TRANSFORMEDDATA
{
	TRANSFORMEDDATA()  { faces = 0; numFaces = 0; }
	~TRANSFORMEDDATA() { if(faces) { delete[] faces; faces = 0; } }
	int numFaces;
	
	POLYGON* faces;
};

struct SCANLINE
{
	int x;						// x coordinate of scanline
	SCALAR z;					// z buffer value
};

struct GSCANLINE
{ 
	int x;
	SCALAR z, color;
};

//	Device3D is the main class that manages all 3d rendering.
//	It holds everything that will be needed to render an object
//	and display it on the screen.
class Device3D
{
public:
	Device3D();						// default constructor
	~Device3D();					// default destructor

	// before device can be used Init() has to called
	bool Init(HWND hwnd);
	void Destroy();

	void BeginRender();				// has to be called before the device can
									// render any object
									// clears the zbuffer and the surface memory;
	void Render(OBJECT3D*);			// renders an object into offscreen memory
	void DrawFrame();				// draws its offscreen memory to the screen

	// the rect defining the screen is changed so
	// the zbuffer and offscreen memory would have
	// to be reset
	bool Update();
	bool SetSurfaceRect(RECT& rc);
	void GetSurfaceRect(RECT& rc)		{ rc = rSrc; }

	// the texture going to be used for texturing
	void SetTexture(TEXTURE3D* Tex)			{ tex = Tex; }
	void GetTexture(TEXTURE3D** Tex)		{ (*Tex) = tex; }		

	// the current shading mode
	void SetShadingMode(DWORD mode)			{ shade_mode = mode; }
	void GetShadingMode(DWORD& mode)		{ mode = shade_mode; }

	// the current rendering mode
	void SetRenderMode(DWORD mode)			{ render_mode = mode; }
	void GetRenderMode(DWORD& mode)			{ mode = render_mode; }

	// the light types
	void SetLight(DWORD lt, LIGHT3D*);
	void GetLight(DWORD lt, LIGHT3D**);
	void RemoveLight(DWORD);

	// the matrix that transforms any object
	// that are rendered by this class
	void SetWorldMatrix(MATRIX& w);
	void GetWorldMatrix(MATRIX& w)		{ w = world; }	
	
	// the view (or camera) matrix
	void SetViewMatrix(VECTOR& vFrom, VECTOR& vLookAt, VECTOR& vWorldUp);
	void SetViewMatrix(MATRIX&);
	void GetViewMatrix(MATRIX& v)		{ v = view; }

	// the projection matrix used to convert
	// 3d coordinates to 2D
	void SetProjectMatrix(MATRIX& p);
	void GetProjectMatrix(MATRIX& p) { p = proj; } 

	// the material tha the object rendered
	// is going to be colored with
	void SetMaterial(MATERIAL3D* m);
	void GetMaterial(MATERIAL3D** m)	{ *m = mat; }

	int dwWidth, dwHeight;			// dimensions of the window
	int clipWidth, clipHeight;		// dimensions of the clipping window
	
private:
	void RenderLineStrip(OBJECT3D*);
	void RenderWireframe(TRANSFORMEDDATA*);	// renders an object's edges	
	void RenderFlat(TRANSFORMEDDATA*);		// renders the object with flat shading
	void RenderGouraud(TRANSFORMEDDATA*);	// renders the object with smooth interpolated shading
	void RenderPhong(TRANSFORMEDDATA*);		// renders the object with phong shading
	
	void ShowStats(HDC&);

	void Clip_Near(POLYGON*, VERTEX3D&, VERTEX3D&);
	void Clip_Left(POLYGON*, VERTEX3D&, VERTEX3D&);
	void Clip_Right(POLYGON*, VERTEX3D&, VERTEX3D&);
	void Clip_Top(POLYGON*, VERTEX3D&, VERTEX3D&);
	void Clip_Bottom(POLYGON*, VERTEX3D&, VERTEX3D&);

	TRANSFORMEDDATA* ComputeScreenCoords(OBJECT3D*);		// computes the screen coords for a give object
	void FillWireframeEdge(LPBYTE*, POINT2D&, POINT2D&);	// draws an edge to the screen
	void FillFlatEdge(SCANLINE*, POINT2D&, POINT2D&);		// fills an edge array with the points that make up a line
	void FillGouraudEdge(GSCANLINE*, POINT2D&, POINT2D&);

	HWND hWnd;						// handle to the window this device is drawing in
	bool isInit;					// is the window initiated with the handle of the window
	RECT rSrc;						// rectangle that defines the screen that it is rendering
	long dwPitch;					// the amount of memory one scanline takes up
	BYTE dwBPP;						// bytes per pixel
	DWORD shade_mode;				// current shading mode
	DWORD render_mode;				// the current rendering mode

	DWORD light_types;				// lights that this object has (can be more than one)
	LIGHT3D *l3dAmbient;				// current lights
	LIGHT3D *l3dDir;			
	LIGHT3D *l3dPoint;			
	LIGHT3D *l3dSpot;		
	VECTOR vtDir, vtPoint;			// transformed light vectors
	VECTOR vtSpotDir, vtSpotPoint;

	MATERIAL3D* mat;				// current material that an object is being colored by
	COLOR3D kAmb;					// coefficients in the lighting illumination model
	COLOR3D kDifDir;
	COLOR3D kDifPoint;
	COLOR3D kDifSpot;

	TEXTURE3D* tex;					// current texture used for texture mapping
									// if 0, then no texture mapping
	MATRIX world;					// world matrix that an object render is transformed by
	MATRIX view;					// matrix storing the view transformation
	MATRIX proj;					// projection matrix that project the 3d coordinates
									// into 2D coordinates
	SCALAR fFarPlane;				// the back clipping plane
	SCALAR fNearPlane;				// the near clipping plane
	
	MATRIX invworld;				// inverse matrix used to transform lights and camera to object space			
	VECTOR vCamPos;					// camera position in model coordinates computed by invview * invworld

	SCALAR* zbuffer;				// used for z-buffering
	HBITMAP hSurface;				// handle to the offscreen surface the device is rendering
	LPBYTE surface;					// memory that holds color data for the surface
	LPBYTE mem;						// a pointer to the memory where the surface resides
									// in such a location that an (x, y) coord can be
									// accessed by mem +	dwPitch*y + x * dwBPP
	DWORD srcMemory;				// the size in bytes of the offscreen surface

};

// Function prototypes
int TestPointInObject(Device3D* dev, OBJECT3D* obj, int sx, int sy);

#endif